#include <elcore50-matrix-lib/convolution_small_asm_ker.h>
#include <elcore50-matrix-lib/convolution_with_dma_fl16.h>
#include <elcore50-matrix-lib/winograd_fl16.h>

extern "C" void float16_to_float32(uint16_t* src, float* dst, int size);
extern "C" void float32_to_float16(float* src, uint16_t* dst, int size);

static void weight_dyxc_to_yxcd(const float* input, float* output, int kernelY, int kernelX, int srcC, int dstC);

#define TEST_COUNT 21

int main()
{
    uint32_t l2dc_val = *L2DC;
    flush_all_caches();
    *L2DC &= 0xFFFFFFBF;

    int tic[2], instr[2];
    Store_version st_ver = STORE_NONE;

    int array_srcH[TEST_COUNT] = { 14,  30,  30,  58,  58,  14,  30,  58,  14,  30, 114, 226, 58, 150,  74,
                                   10,  10,  34,  34,  34,  66};

    int array_srcW[TEST_COUNT] = { 14,  30,  30,  58,  58,  14,  30,  58,  14,  30, 114, 226, 58, 150,  74,
                                   34,  34,  34,  34,  34,  66};

    int array_srcC[TEST_COUNT] = {128, 128, 128, 128, 128, 256, 256, 256, 512, 512,  64,  64, 64,  32,  96,
                                  192, 256, 256, 512, 512, 512};

    int array_dstC[TEST_COUNT] = { 32, 128,  32, 256,  32, 256, 512, 256, 512, 512, 128,  64, 64,  32, 192,
                                  256, 256, 256, 256, 512, 512};

    printf("| func_name        | input (b, h, w, ch)    | bias | store mode | conv2d mul/tic | conv_tic/win_tic | "
           "status |\n");
    printf("|                  | ker (X, Y), str (X, Y) |      |            |                |                  | "
           "       |\n");
    printf("|                  | output (b, h, w, ch)   |      |            |                |                  | "
           "       |\n");
    printf("----------------------------------------------------------------------------------------------------"
           "----------\n");

    for (int test_id = 0; test_id < TEST_COUNT; ++test_id) {
      int srcH = array_srcH[test_id];
      int srcW = array_srcW[test_id];
      int srcC = array_srcC[test_id];
      int dstC = array_dstC[test_id];

      printf("| winograd vs conv | (%1d %4d %4d %4d)     |    %d |", 1, srcH, srcW, srcC, 0);
      printf("      none  | ");

      int dstH = srcH - 2;
      int dstW = srcW - 2;

      int kernelY = 3;
      int kernelX = 3;

      float* src = (float*)memalign(64, srcH * srcW * srcC * sizeof(float));
      float* weight_dyxc = (float*)memalign(64, kernelY * kernelX * dstC * srcC * sizeof(float));
      float* weight_yxcd = (float*)memalign(64, kernelY * kernelX * dstC * srcC * sizeof(float));

      for (int i = 0; i < srcH * srcW * srcC; ++i) {
        src[i] = rand() % 3 - 1;
      }

      for (int i = 0; i < kernelY * kernelX * dstC * srcC; ++i) {
        weight_dyxc[i] = rand() % 3 - 1;
      }

      weight_dyxc_to_yxcd(weight_dyxc, weight_yxcd, kernelY, kernelX, srcC, dstC);

      uint16_t* weight_yxcd_fl16 = (uint16_t*)memalign(64, kernelY * kernelX * dstC * srcC * sizeof(uint16_t));
      float32_to_float16(weight_yxcd, weight_yxcd_fl16, kernelY * kernelX * dstC * srcC);

      float* dst_conv2d = (float*)memalign(64, dstC * dstH * dstW * sizeof(float));
      float* dst_winograd = (float*)memalign(64, dstC * dstH * dstW * sizeof(float));

      uint16_t* src_fl16 = (uint16_t*)memalign(64, srcH * srcW * srcC * sizeof(uint16_t));
      uint16_t* dst_conv2d_fl16 = (uint16_t*)memalign(64, dstC * dstH * dstW * sizeof(uint16_t));
      uint16_t* dst_winograd_fl16 = (uint16_t*)memalign(64, dstC * dstH * dstW * sizeof(uint16_t));
      float32_to_float16(src, src_fl16, srcH * srcW * srcC);

      /* Conv2d */
      ConvLocalMemBufFl16 local_buf;
      Debug_data tile_params = { 4,   4,   4,  4,  8,  128};
      VDMAChain chain_A, chain_B, chain_st_C, chain_ld_C, chain_init_vec;
      MatMulDmaChains4 chains = {&chain_A, &chain_B, &chain_st_C, &chain_ld_C, &chain_init_vec};

      int initcnt = 32;
      initchain(chains.chain_A, initcnt);
      initchain(chains.chain_B, initcnt);
      initchain(chains.chain_st_C, initcnt);
      initchain(chains.chain_ld_C, initcnt);
      initchain(chains.chain_init_vec, initcnt);

      uint16_t** zeros = (uint16_t**)malloc(sizeof(uint16_t*));

      conv_size_selector(1, srcH, srcW, srcC, kernelY, kernelX, dstH, dstW, dstC, &local_buf,
                         1, 1, 2, 2, 2, 0, XYRAM_SIZE, &(local_buf.offset_A));

      Conv_version conv_ver;
      if (local_buf.buf_dstC > 64)
      {
          conv_ver = VER_4x4x128;
      }
      else if (local_buf.buf_dstC > 32)
      {
          conv_ver = VER_4x8x64;
      }
      else
      {
          conv_ver = VER_8x8x32;
      }


      if (local_buf.buf_srcC >= srcC)
      {
          /* версия без шага по srcC */
          init_chain_dma_convolution_small_ker_2A2B2C(src_fl16, 1, srcH, srcW, srcC,
                                                     kernelY, kernelX, 1, 1,
                                                     1, 1, dstH, dstW, 1,
                                                     0, 0, 0, 0,
                                                     weight_yxcd_fl16, NULL, dst_conv2d_fl16, dstC, &local_buf, &chains,
                                                     zeros, 0, (uint16_t*)&__local_mem, XYRAM_SIZE);
          count_tics(tic, instr);
          convolution_small_ker_2A2B2C(src_fl16, 1, srcH, srcW, srcC,
                                      kernelY, kernelX, 1, 1,
                                      1, 1, dstH, dstW, 1,
                                      weight_yxcd_fl16, NULL, dst_conv2d_fl16, dstC,
                                      &local_buf, &chains, NULL, st_ver, 0, conv_ver);

          count_tics(&tic[1], &instr[1]);
      }
      else
      {
          /* версия с шагом по srcC */
          float** inter_res = (float**)malloc(sizeof(float*));

          init_chain_dma_convolution_small_ker_2A2B2C_srcC_ver(src_fl16, 1, srcH, srcW, srcC,
                                                     kernelY, kernelX, 1, 1,
                                                     1, 1, dstH, dstW, 1,
                                                     0, 0, 0, 0,
                                                     weight_yxcd_fl16, NULL, dst_conv2d_fl16, dstC, &local_buf, &chains,
                                                     zeros, 0, (uint16_t*)&__local_mem, 524288, inter_res,
                                                     tile_params);

          count_tics(tic, instr);
          convolution_small_ker_2A2B2C_srcC_ver(src_fl16, 1, srcH, srcW, srcC,
                                      kernelY, kernelX, 1, 1,
                                      1, 1, dstH, dstW, 1,
                                      weight_yxcd_fl16, NULL, dst_conv2d_fl16, dstC,
                                      &local_buf, &chains, NULL, st_ver, 0, *inter_res, conv_ver);

          count_tics(&tic[1], &instr[1]);
          free(*inter_res);
          free(inter_res);
      }

      destroychain(chains.chain_A);
      destroychain(chains.chain_B);
      destroychain(chains.chain_st_C);
      destroychain(chains.chain_ld_C);
      destroychain(chains.chain_init_vec);

      flush_all_caches();

      float16_to_float32(dst_conv2d_fl16, dst_conv2d, dstC * dstH * dstW);
      int conv_tic = tic[1] - tic[0];
      printf("    %9.3f  | ", (float)dstH * dstW * srcC * 3 * 3 * dstC / (tic[1] - tic[0]));

      free(*zeros);
      free(zeros);

      /* Winograd 2x2 */
      WinogradConfig config;
      init_dma_chain_winograd(
        src_fl16, 1, srcC, srcH, srcW, weight_yxcd, dst_winograd_fl16, dstC, &config);

      count_tics(tic, instr);
      run_winograd(src_fl16, 1, srcC, srcH, srcW, dst_winograd_fl16, dstC, &config);
      count_tics(&tic[1], &instr[1]);

      printf("     %10.3f  | ", (float)conv_tic / (tic[1] - tic[0]));

      float16_to_float32(dst_winograd_fl16, dst_winograd, dstC * dstH * dstW);
      destroy_dma_chain_winograd(&config);
      free(src_fl16);
      free(dst_conv2d_fl16);
      free(dst_winograd_fl16);

      int error = 0;
      for (int i = 0; i < dstH * dstW * dstC; ++i) {
        if (dst_conv2d[i] != dst_winograd[i]) {
              error++;
              break;
        }
      }

      free(src);
      free(weight_dyxc);
      free(weight_yxcd);
      free(dst_conv2d);
      free(dst_winograd);
      free(weight_yxcd_fl16);

      if (!error)
          printf("passed |\n");
      else
          printf("failed |\n");

      printf("|                  | (%1d %2d) (%1d %2d)          |      |            |                |                 "
             " |        |\n", 3, 3, 1, 1);
      printf("|                  | (%1d %4d %4d %4d)     |      |            |                |                  |     "
             "   |\n", 1, dstH, dstW, dstC);
      printf("---------------------------------------------------------------------------------------------------------"
             "-----\n");
    }

    *L2DC = l2dc_val;

    return 0;
}

__attribute__((optnone))
static void weight_dyxc_to_yxcd(const float* input, float* output, int kernelY, int kernelX, int srcC, int dstC)
{
    int indx_in = 0;
    int indx_out = 0;
    for (int h = 0; h < kernelY; ++h)
    {
        for (int w = 0; w < kernelX; ++w)
        {
            for (int ch = 0; ch < srcC; ++ch)
            {
                for (int b = 0; b < dstC; ++b)
                {
                    indx_out = h * dstC * kernelX * srcC
                               + w * srcC * dstC
                               + ch * dstC + b;
                    indx_in = b * kernelY * kernelX * srcC
                              + h * kernelX * srcC
                              + w * srcC + ch;
                    output[indx_out] = input[indx_in];
                }
            }
        }
    }
}
